import time
import warnings
from datetime import timedelta as TimeDelta
from typing import Any, Optional, Protocol

from flask import Flask

from .._utils import total_seconds
from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class MemcacheClientProtocol(Protocol):
    def get(self, key: str) -> Optional[Any]: ...
    def set(self, key: str, value: Any, timeout: int) -> bool: ...
    def delete(self, key: str) -> bool: ...


class MemcachedSession(ServerSideSession):
    pass


class MemcachedSessionInterface(ServerSideSessionInterface):
    """A Session interface that uses memcached as session storage. (`pylibmc`, `libmc`, `python-memcached` or `pymemcache` required)

    :param client: A ``memcache.Client`` instance.
    :param key_prefix: A prefix that is added to all storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.

    .. versionadded:: 0.7
        The `serialization_format` and `app` parameters were added.

    .. versionadded:: 0.6
        The `sid_length` parameter was added.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.
    """

    serializer = ServerSideSessionInterface.serializer
    session_class = MemcachedSession
    ttl = True

    def __init__(
        self,
        app: Flask,
        client: Optional[MemcacheClientProtocol] = Defaults.SESSION_MEMCACHED,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
    ):
        if client is None or not all(
            hasattr(client, method) for method in ["get", "set", "delete"]
        ):
            warnings.warn(
                "No valid memcache.Client instance provided, attempting to create a new instance on localhost with default settings.",
                RuntimeWarning,
                stacklevel=1,
            )
            client = self._get_preferred_memcache_client()
        self.client = client
        super().__init__(
            app, key_prefix, use_signer, permanent, sid_length, serialization_format
        )

    def _get_preferred_memcache_client(self):
        clients = [
            ("pylibmc", ["127.0.0.1:11211"]),
            ("memcache", ["127.0.0.1:11211"]),  # python-memcached
            ("pymemcache.client.base", "127.0.0.1:11211"),
            ("libmc", ["localhost:11211"]),
        ]

        for module_name, server in clients:
            try:
                module = __import__(module_name)
                ClientClass = module.Client
                return ClientClass(server)
            except ImportError:
                continue

        raise ImportError("No memcache module found")

    def _get_memcache_timeout(self, timeout: int) -> int:
        """
        Memcached deals with long (> 30 days) timeouts in a special
        way. Call this function to obtain a safe value for your timeout.
        """
        if timeout > 2592000:  # 60*60*24*30, 30 days
            # Switch to absolute timestamps.
            timeout += int(time.time())
        return timeout

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (item) from the database
        serialized_session_data = self.client.get(store_id)
        if serialized_session_data:
            return self.serializer.decode(serialized_session_data)
        return None

    def _delete_session(self, store_id: str) -> None:
        self.client.delete(store_id)

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_time_to_live = total_seconds(session_lifetime)

        # Serialize the session data
        serialized_session_data = self.serializer.encode(session)

        # Update existing or create new session in the database
        self.client.set(
            store_id,
            serialized_session_data,
            self._get_memcache_timeout(storage_time_to_live),
        )



import warnings
from datetime import datetime
from datetime import timedelta as TimeDelta
from typing import Optional

from flask import Flask
from itsdangerous import want_bytes
from pymongo import MongoClient, version

from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class MongoDBSession(ServerSideSession):
    pass


class MongoDBSessionInterface(ServerSideSessionInterface):
    """A Session interface that uses mongodb as session storage. (`pymongo` required)

    :param client: A ``pymongo.MongoClient`` instance.
    :param key_prefix: A prefix that is added to all storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.
    :param db: The database to use.
    :param collection: The collection to use.

    .. versionadded:: 0.7
        The `serialization_format` and `app` parameters were added.

    .. versionadded:: 0.6
        The `sid_length` parameter was added.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.
    """

    session_class = MongoDBSession
    ttl = True

    def __init__(
        self,
        app: Flask,
        client: Optional[MongoClient] = Defaults.SESSION_MONGODB,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
        db: str = Defaults.SESSION_MONGODB_DB,
        collection: str = Defaults.SESSION_MONGODB_COLLECT,
    ):

        if client is None or not isinstance(client, MongoClient):
            warnings.warn(
                "No valid MongoClient instance provided, attempting to create a new instance on localhost with default settings.",
                RuntimeWarning,
                stacklevel=1,
            )
            client = MongoClient()

        self.client = client
        self.store = client[db][collection]
        self.use_deprecated_method = int(version.split(".")[0]) < 4

        # Create a TTL index on the expiration time, so that mongo can automatically delete expired sessions
        self.store.create_index("expiration", expireAfterSeconds=0)

        super().__init__(
            app, key_prefix, use_signer, permanent, sid_length, serialization_format
        )

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (document) from the database
        document = self.store.find_one({"id": store_id})
        if document:
            serialized_session_data = want_bytes(document["val"])
            return self.serializer.decode(serialized_session_data)
        return None

    def _delete_session(self, store_id: str) -> None:
        if self.use_deprecated_method:
            self.store.remove({"id": store_id})
        else:
            self.store.delete_one({"id": store_id})

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_expiration_datetime = datetime.utcnow() + session_lifetime

        # Serialize the session data
        serialized_session_data = self.serializer.encode(session)

        # Update existing or create new session in the database
        if self.use_deprecated_method:
            self.store.update(
                {"id": store_id},
                {
                    "id": store_id,
                    "val": serialized_session_data,
                    "expiration": storage_expiration_datetime,
                },
                True,
            )
        else:
            self.store.update_one(
                {"id": store_id},
                {
                    "$set": {
                        "id": store_id,
                        "val": serialized_session_data,
                        "expiration": storage_expiration_datetime,
                    }
                },
                True,
            )



import warnings
from datetime import timedelta as TimeDelta
from typing import Optional

from cachelib.file import FileSystemCache
from flask import Flask

from .._utils import total_seconds
from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class FileSystemSession(ServerSideSession):
    pass


class FileSystemSessionInterface(ServerSideSessionInterface):
    """Uses the :class:`cachelib.file.FileSystemCache` as a session storage.

    :param key_prefix: A prefix that is added to storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.
    :param cache_dir: the directory where session files are stored.
    :param threshold: the maximum number of items the session stores before it
    :param mode: the file mode wanted for the session files, default 0600

    .. versionadded:: 0.7
        The `serialization_format` and `app` parameters were added.

    .. versionadded:: 0.6
        The `sid_length` parameter was added.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.
    """

    session_class = FileSystemSession
    ttl = True

    def __init__(
        self,
        app: Flask,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
        cache_dir: str = Defaults.SESSION_FILE_DIR,
        threshold: int = Defaults.SESSION_FILE_THRESHOLD,
        mode: int = Defaults.SESSION_FILE_MODE,
    ):

        # Deprecation warnings
        if cache_dir != Defaults.SESSION_FILE_DIR:
            warnings.warn(
                "'SESSION_FILE_DIR' is deprecated and will be removed in a future release. Instead pass FileSystemCache(directory, threshold, mode) instance as SESSION_CACHELIB.",
                DeprecationWarning,
                stacklevel=1,
            )
        if threshold != Defaults.SESSION_FILE_THRESHOLD:
            warnings.warn(
                "'SESSION_FILE_THRESHOLD' is deprecated and will be removed in a future release. Instead pass FileSystemCache(directory, threshold, mode) instance as SESSION_CLIENT.",
                DeprecationWarning,
                stacklevel=1,
            )
        if mode != Defaults.SESSION_FILE_MODE:
            warnings.warn(
                "'SESSION_FILE_MODE' is deprecated and will be removed in a future release. Instead pass FileSystemCache(directory, threshold, mode) instance as SESSION_CLIENT.",
                DeprecationWarning,
                stacklevel=1,
            )

        warnings.warn(
            "FileSystemSessionInterface is deprecated and will be removed in a future release. Instead use the CacheLib backend directly.",
            DeprecationWarning,
            stacklevel=1,
        )

        self.cache = FileSystemCache(
            cache_dir=cache_dir, threshold=threshold, mode=mode
        )

        super().__init__(
            app, key_prefix, use_signer, permanent, sid_length, serialization_format
        )

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (item) from the database
        return self.cache.get(store_id)

    def _delete_session(self, store_id: str) -> None:
        self.cache.delete(store_id)

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_time_to_live = total_seconds(session_lifetime)

        # Serialize the session data (or just cast into dictionary in this case)
        session_data = dict(session)

        # Update existing or create new session in the database
        self.cache.set(
            key=store_id,
            value=session_data,
            timeout=storage_time_to_live,
        )



import warnings
from datetime import datetime
from datetime import timedelta as TimeDelta
from typing import Any, Optional

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from itsdangerous import want_bytes
from sqlalchemy import Column, DateTime, Integer, LargeBinary, Sequence, String

from .._utils import retry_query
from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class SqlAlchemySession(ServerSideSession):
    pass


def create_session_model(db, table_name, schema=None, bind_key=None, sequence=None):
    class Session(db.Model):
        __tablename__ = table_name
        __table_args__ = {"schema": schema} if schema else {}
        __bind_key__ = bind_key

        id = (
            Column(Integer, Sequence(sequence), primary_key=True)
            if sequence
            else Column(Integer, primary_key=True)
        )
        session_id = Column(String(255), unique=True)
        data = Column(LargeBinary)
        expiry = Column(DateTime)

        def __init__(self, session_id: str, data: Any, expiry: datetime):
            self.session_id = session_id
            self.data = data
            self.expiry = expiry

        def __repr__(self):
            return f"<Session data {self.data}>"

    return Session


class SqlAlchemySessionInterface(ServerSideSessionInterface):
    """Uses the Flask-SQLAlchemy from a flask app as session storage.

    :param app: A Flask app instance.
    :param client: A Flask-SQLAlchemy instance.
    :param key_prefix: A prefix that is added to all storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.
    :param table: The table name you want to use.
    :param sequence: The sequence to use for the primary key if needed.
    :param schema: The db schema to use.
    :param bind_key: The db bind key to use.
    :param cleanup_n_requests: Delete expired sessions on average every N requests.

    .. versionadded:: 0.7
        db changed to client to be standard on all session interfaces.
        The `cleanup_n_request` parameter was added.

    .. versionadded:: 0.6
        The `sid_length`, `sequence`, `schema` and `bind_key` parameters were added.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.
    """

    session_class = SqlAlchemySession
    ttl = False

    def __init__(
        self,
        app: Optional[Flask],
        client: Optional[SQLAlchemy] = Defaults.SESSION_SQLALCHEMY,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
        table: str = Defaults.SESSION_SQLALCHEMY_TABLE,
        sequence: Optional[str] = Defaults.SESSION_SQLALCHEMY_SEQUENCE,
        schema: Optional[str] = Defaults.SESSION_SQLALCHEMY_SCHEMA,
        bind_key: Optional[str] = Defaults.SESSION_SQLALCHEMY_BIND_KEY,
        cleanup_n_requests: Optional[int] = Defaults.SESSION_CLEANUP_N_REQUESTS,
    ):
        self.app = app

        if client is None or not isinstance(client, SQLAlchemy):
            warnings.warn(
                "No valid SQLAlchemy instance provided, attempting to create a new instance on localhost with default settings.",
                RuntimeWarning,
                stacklevel=1,
            )
            client = SQLAlchemy(app)
        self.client = client

        # Create the session model
        self.sql_session_model = create_session_model(
            client, table, schema, bind_key, sequence
        )
        # Create the table if it does not exist
        with app.app_context():
            if bind_key:
                engine = self.client.get_engine(app, bind=bind_key)
            else:
                engine = self.client.engine
            self.sql_session_model.__table__.create(bind=engine, checkfirst=True)

        super().__init__(
            app,
            key_prefix,
            use_signer,
            permanent,
            sid_length,
            serialization_format,
            cleanup_n_requests,
        )

    @retry_query()
    def _delete_expired_sessions(self) -> None:
        try:
            self.client.session.query(self.sql_session_model).filter(
                self.sql_session_model.expiry <= datetime.utcnow()
            ).delete(synchronize_session=False)
            self.client.session.commit()
        except Exception:
            self.client.session.rollback()
            raise

    @retry_query()
    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (record) from the database
        record = self.sql_session_model.query.filter_by(session_id=store_id).first()

        # "Delete the session record if it is expired as SQL has no TTL ability
        if record and (record.expiry is None or record.expiry <= datetime.utcnow()):
            try:
                self.client.session.delete(record)
                self.client.session.commit()
            except Exception:
                self.client.session.rollback()
                raise
            record = None

        if record:
            serialized_session_data = want_bytes(record.data)
            return self.serializer.decode(serialized_session_data)
        return None

    @retry_query()
    def _delete_session(self, store_id: str) -> None:
        try:
            self.sql_session_model.query.filter_by(session_id=store_id).delete()
            self.client.session.commit()
        except Exception:
            self.client.session.rollback()
            raise

    @retry_query()
    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_expiration_datetime = datetime.utcnow() + session_lifetime

        # Serialize session data
        serialized_session_data = self.serializer.encode(session)

        # Update existing or create new session in the database
        try:
            record = self.sql_session_model.query.filter_by(session_id=store_id).first()
            if record:
                record.data = serialized_session_data
                record.expiry = storage_expiration_datetime
            else:
                record = self.sql_session_model(
                    session_id=store_id,
                    data=serialized_session_data,
                    expiry=storage_expiration_datetime,
                )
                self.client.session.add(record)
            self.client.session.commit()
        except Exception:
            self.client.session.rollback()
            raise



import warnings
from datetime import timedelta as TimeDelta
from typing import Optional

from flask import Flask
from redis import Redis

from .._utils import total_seconds
from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class RedisSession(ServerSideSession):
    pass


class RedisSessionInterface(ServerSideSessionInterface):
    """Uses the Redis key-value store as a session storage. (`redis-py` required)

    :param client: A ``redis.Redis`` instance.
    :param key_prefix: A prefix that is added to all storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.

    .. versionadded:: 0.7
        The `serialization_format` and `app` parameters were added.

    .. versionadded:: 0.6
        The `sid_length` parameter was added.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.
    """

    session_class = RedisSession
    ttl = True

    def __init__(
        self,
        app: Flask,
        client: Optional[Redis] = Defaults.SESSION_REDIS,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
    ):
        if client is None or not isinstance(client, Redis):
            warnings.warn(
                "No valid Redis instance provided, attempting to create a new instance on localhost with default settings.",
                RuntimeWarning,
                stacklevel=1,
            )
            client = Redis()
        self.client = client
        super().__init__(
            app, key_prefix, use_signer, permanent, sid_length, serialization_format
        )

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (value) from the database
        serialized_session_data = self.client.get(store_id)
        if serialized_session_data:
            return self.serializer.decode(serialized_session_data)
        return None

    def _delete_session(self, store_id: str) -> None:
        self.client.delete(store_id)

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_time_to_live = total_seconds(session_lifetime)

        # Serialize the session data
        serialized_session_data = self.serializer.encode(session)

        # Update existing or create new session in the database
        self.client.set(
            name=store_id,
            value=serialized_session_data,
            ex=storage_time_to_live,
        )



import warnings
from datetime import datetime
from datetime import timedelta as TimeDelta
from decimal import Decimal
from typing import Optional

import boto3
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource
from flask import Flask
from itsdangerous import want_bytes

from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class DynamoDBSession(ServerSideSession):
    pass


class DynamoDBSessionInterface(ServerSideSessionInterface):
    """A Session interface that uses dynamodb as backend. (`boto3` required)

    :param client: A ``DynamoDBServiceResource`` instance.
    :param key_prefix: A prefix that is added to all DynamoDB store keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param table_name: DynamoDB table name to store the session.

    .. versionadded:: 0.6
        The `sid_length` parameter was added.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.
    """

    session_class = DynamoDBSession

    def __init__(
        self,
        app: Flask,
        client: Optional[DynamoDBServiceResource] = Defaults.SESSION_DYNAMODB,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
        table_name: str = Defaults.SESSION_DYNAMODB_TABLE,
    ):

        if client is None:
            warnings.warn(
                "No valid DynamoDBServiceResource instance provided, attempting to create a new instance on localhost:8000.",
                RuntimeWarning,
                stacklevel=1,
            )
            client = boto3.resource(
                "dynamodb",
                endpoint_url="http://localhost:8000",
                region_name="us-west-2",
                aws_access_key_id="dummy",
                aws_secret_access_key="dummy",
            )

        try:
            client.create_table(
                AttributeDefinitions=[
                    {"AttributeName": "id", "AttributeType": "S"},
                ],
                TableName=table_name,
                KeySchema=[
                    {"AttributeName": "id", "KeyType": "HASH"},
                ],
                BillingMode="PAY_PER_REQUEST",
            )
            client.meta.client.get_waiter("table_exists").wait(TableName=table_name)
            client.meta.client.update_time_to_live(
                TableName=self.table_name,
                TimeToLiveSpecification={
                    "Enabled": True,
                    "AttributeName": "expiration",
                },
            )
        except (AttributeError, client.meta.client.exceptions.ResourceInUseException):
            # TTL already exists, or table already exists
            pass

        self.client = client
        self.store = client.Table(table_name)
        super().__init__(
            app,
            key_prefix,
            use_signer,
            permanent,
            sid_length,
            serialization_format,
        )

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (document) from the database
        document = self.store.get_item(Key={"id": store_id}).get("Item")
        if document:
            serialized_session_data = want_bytes(document.get("val").value)
            return self.serializer.decode(serialized_session_data)
        return None

    def _delete_session(self, store_id: str) -> None:
        self.store.delete_item(Key={"id": store_id})

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_expiration_datetime = datetime.utcnow() + session_lifetime
        # Serialize the session data
        serialized_session_data = self.serializer.encode(session)

        self.store.update_item(
            Key={
                "id": store_id,
            },
            UpdateExpression="SET val = :value, expiration = :exp",
            ExpressionAttributeValues={
                ":value": serialized_session_data,
                ":exp": Decimal(storage_expiration_datetime.timestamp()),
            },
        )



import warnings
from datetime import timedelta as TimeDelta
from typing import Optional

from cachelib.file import FileSystemCache
from flask import Flask

from .._utils import total_seconds
from ..base import ServerSideSession, ServerSideSessionInterface
from ..defaults import Defaults


class CacheLibSession(ServerSideSession):
    pass


class CacheLibSessionInterface(ServerSideSessionInterface):
    """Uses any :class:`cachelib` backend as a session storage.

    :param client: A :class:`cachelib` backend instance.
    :param key_prefix: A prefix that is added to storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.
    """

    session_class = CacheLibSession
    ttl = True

    def __init__(
        self,
        app: Flask = None,
        client: Optional[FileSystemCache] = Defaults.SESSION_CACHELIB,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
    ):

        if client is None:
            warnings.warn(
                "No valid cachelib instance provided, attempting to create a new instance on localhost with default settings.",
                RuntimeWarning,
                stacklevel=1,
            )
            client = FileSystemCache("flask_session", threshold=500)

        self.cache = client

        super().__init__(
            None, key_prefix, use_signer, permanent, sid_length, serialization_format
        )

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (item) from the database
        return self.cache.get(store_id)

    def _delete_session(self, store_id: str) -> None:
        self.cache.delete(store_id)

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_time_to_live = total_seconds(session_lifetime)

        # Serialize the session data (or just cast into dictionary in this case)
        session_data = dict(session)

        # Update existing or create new session in the database
        self.cache.set(
            key=store_id,
            value=session_data,
            timeout=storage_time_to_live,
        )



from .defaults import Defaults

__version__ = "0.8.0"


class Session:
    """This class is used to add Server-side Session to one or more Flask
    applications.

    :param app: A Flask app instance.

    For a typical setup use the following initialization::

        app = Flask(__name__)
        Session(app)

    .. note::

        You can not use ``Session`` instance directly, what ``Session`` does
        is just change the :attr:`~flask.Flask.session_interface` attribute on
        your Flask applications. You should always use :class:`flask.session`.
    """

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """This the the alternate setup method, typically used in an application factory pattern::

            sess = Session()

            def create_app():
                app = Flask(__name__)
                sess.init_app(app)
                return app

        :param app: the Flask app object with proper configuration.
        """
        app.session_interface = self._get_interface(app)

    def _get_interface(self, app):
        config = app.config

        # Flask-session specific settings
        SESSION_TYPE = config.get("SESSION_TYPE", Defaults.SESSION_TYPE)

        SESSION_PERMANENT = config.get("SESSION_PERMANENT", Defaults.SESSION_PERMANENT)
        SESSION_USE_SIGNER = config.get(
            "SESSION_USE_SIGNER", Defaults.SESSION_USE_SIGNER
        )  # TODO: remove in 1.0
        SESSION_KEY_PREFIX = config.get(
            "SESSION_KEY_PREFIX", Defaults.SESSION_KEY_PREFIX
        )
        SESSION_ID_LENGTH = config.get("SESSION_ID_LENGTH", Defaults.SESSION_ID_LENGTH)
        SESSION_SERIALIZATION_FORMAT = config.get(
            "SESSION_SERIALIZATION_FORMAT", Defaults.SESSION_SERIALIZATION_FORMAT
        )

        # Redis settings
        SESSION_REDIS = config.get("SESSION_REDIS", Defaults.SESSION_REDIS)

        # Memcached settings
        SESSION_MEMCACHED = config.get("SESSION_MEMCACHED", Defaults.SESSION_MEMCACHED)

        # CacheLib settings
        SESSION_CACHELIB = config.get("SESSION_CACHELIB", Defaults.SESSION_CACHELIB)

        # Filesystem settings
        # TODO: remove in 1.0
        SESSION_FILE_DIR = config.get("SESSION_FILE_DIR", Defaults.SESSION_FILE_DIR)
        SESSION_FILE_THRESHOLD = config.get(
            "SESSION_FILE_THRESHOLD", Defaults.SESSION_FILE_THRESHOLD
        )
        SESSION_FILE_MODE = config.get("SESSION_FILE_MODE", Defaults.SESSION_FILE_MODE)

        # MongoDB settings
        SESSION_MONGODB = config.get("SESSION_MONGODB", Defaults.SESSION_MONGODB)
        SESSION_MONGODB_DB = config.get(
            "SESSION_MONGODB_DB", Defaults.SESSION_MONGODB_DB
        )
        SESSION_MONGODB_COLLECT = config.get(
            "SESSION_MONGODB_COLLECT", Defaults.SESSION_MONGODB_COLLECT
        )

        # SQLAlchemy settings
        SESSION_SQLALCHEMY = config.get(
            "SESSION_SQLALCHEMY", Defaults.SESSION_SQLALCHEMY
        )
        SESSION_SQLALCHEMY_TABLE = config.get(
            "SESSION_SQLALCHEMY_TABLE", Defaults.SESSION_SQLALCHEMY_TABLE
        )
        SESSION_SQLALCHEMY_SEQUENCE = config.get(
            "SESSION_SQLALCHEMY_SEQUENCE", Defaults.SESSION_SQLALCHEMY_SEQUENCE
        )
        SESSION_SQLALCHEMY_SCHEMA = config.get(
            "SESSION_SQLALCHEMY_SCHEMA", Defaults.SESSION_SQLALCHEMY_SCHEMA
        )
        SESSION_SQLALCHEMY_BIND_KEY = config.get(
            "SESSION_SQLALCHEMY_BIND_KEY", Defaults.SESSION_SQLALCHEMY_BIND_KEY
        )
        SESSION_CLEANUP_N_REQUESTS = config.get(
            "SESSION_CLEANUP_N_REQUESTS", Defaults.SESSION_CLEANUP_N_REQUESTS
        )

        # DynamoDB settings
        SESSION_DYNAMODB = config.get("SESSION_DYNAMODB", Defaults.SESSION_DYNAMODB)
        SESSION_DYNAMODB_TABLE = config.get(
            "SESSION_DYNAMODB_TABLE", Defaults.SESSION_DYNAMODB_TABLE
        )

        common_params = {
            "app": app,
            "key_prefix": SESSION_KEY_PREFIX,
            "use_signer": SESSION_USE_SIGNER,
            "permanent": SESSION_PERMANENT,
            "sid_length": SESSION_ID_LENGTH,
            "serialization_format": SESSION_SERIALIZATION_FORMAT,
        }

        SESSION_TYPE = SESSION_TYPE.lower()

        if SESSION_TYPE == "redis":
            from .redis import RedisSessionInterface

            session_interface = RedisSessionInterface(
                **common_params,
                client=SESSION_REDIS,
            )
        elif SESSION_TYPE == "memcached":
            from .memcached import MemcachedSessionInterface

            session_interface = MemcachedSessionInterface(
                **common_params,
                client=SESSION_MEMCACHED,
            )
        elif SESSION_TYPE == "filesystem":
            from .filesystem import FileSystemSessionInterface

            session_interface = FileSystemSessionInterface(
                **common_params,
                cache_dir=SESSION_FILE_DIR,
                threshold=SESSION_FILE_THRESHOLD,
                mode=SESSION_FILE_MODE,
            )
        elif SESSION_TYPE == "cachelib":
            from .cachelib import CacheLibSessionInterface

            session_interface = CacheLibSessionInterface(
                **common_params, client=SESSION_CACHELIB
            )
        elif SESSION_TYPE == "mongodb":
            from .mongodb import MongoDBSessionInterface

            session_interface = MongoDBSessionInterface(
                **common_params,
                client=SESSION_MONGODB,
                db=SESSION_MONGODB_DB,
                collection=SESSION_MONGODB_COLLECT,
            )
        elif SESSION_TYPE == "sqlalchemy":
            from .sqlalchemy import SqlAlchemySessionInterface

            session_interface = SqlAlchemySessionInterface(
                **common_params,
                client=SESSION_SQLALCHEMY,
                table=SESSION_SQLALCHEMY_TABLE,
                sequence=SESSION_SQLALCHEMY_SEQUENCE,
                schema=SESSION_SQLALCHEMY_SCHEMA,
                bind_key=SESSION_SQLALCHEMY_BIND_KEY,
                cleanup_n_requests=SESSION_CLEANUP_N_REQUESTS,
            )
        elif SESSION_TYPE == "dynamodb":
            from .dynamodb import DynamoDBSessionInterface

            session_interface = DynamoDBSessionInterface(
                **common_params,
                client=SESSION_DYNAMODB,
                table_name=SESSION_DYNAMODB_TABLE,
            )

        else:
            raise ValueError(f"Unrecognized value for SESSION_TYPE: {SESSION_TYPE}")

        return session_interface



import os


class Defaults:
    # Flask-session specific settings
    SESSION_TYPE = "null"
    SESSION_KEY_PREFIX = "session:"
    SESSION_USE_SIGNER = False
    SESSION_PERMANENT = True
    SESSION_ID_LENGTH = 32
    SESSION_SERIALIZATION_FORMAT = "msgpack"

    # Clean up settings for non TTL backends (SQL, PostgreSQL, etc.)
    SESSION_CLEANUP_N_REQUESTS = None

    # Redis settings
    SESSION_REDIS = None

    # Memcached settings
    SESSION_MEMCACHED = None

    # CacheLib settings
    SESSION_CACHELIB = None

    # Filesystem settings
    # TODO: remove in 1.0
    SESSION_FILE_DIR = os.path.join(os.getcwd(), "flask_session")
    SESSION_FILE_THRESHOLD = 500
    SESSION_FILE_MODE = 384

    # MongoDB settings
    SESSION_MONGODB = None
    SESSION_MONGODB_DB = "flask_session"
    SESSION_MONGODB_COLLECT = "sessions"

    # SQLAlchemy settings
    SESSION_SQLALCHEMY = None
    SESSION_SQLALCHEMY_TABLE = "sessions"
    SESSION_SQLALCHEMY_SEQUENCE = None
    SESSION_SQLALCHEMY_SCHEMA = None
    SESSION_SQLALCHEMY_BIND_KEY = None

    # DynamoDB settings
    SESSION_DYNAMODB = None
    SESSION_DYNAMODB_TABLE = "Sessions"



import secrets
import warnings
from abc import ABC, abstractmethod
from contextlib import suppress

try:
    import cPickle as pickle
except ImportError:
    import pickle

import random
from datetime import timedelta as TimeDelta
from typing import Any, Dict, Optional

import msgspec
from flask import Flask, Request, Response
from flask.sessions import SessionInterface as FlaskSessionInterface
from flask.sessions import SessionMixin
from itsdangerous import BadSignature, Signer, want_bytes
from werkzeug.datastructures import CallbackDict

from ._utils import retry_query
from .defaults import Defaults


class ServerSideSession(CallbackDict, SessionMixin):
    """Baseclass for server-side based sessions. This can be accessed through ``flask.session``.

    .. attribute:: sid

    Session id, internally we use :func:`secrets.token_urlsafe` to generate one
    session id.

    .. attribute:: modified

    When data is changed, this is set to ``True``. Only the session dictionary
    itself is tracked; if the session contains mutable data (for example a nested
    dict) then this must be set to ``True`` manually when modifying that data. The
    session cookie will only be written to the response if this is ``True``.

    .. attribute:: accessed

    When data is read (or attempted read) or written, this is set to ``True``. Used by
    :class:`.ServerSideSessionInterface` to add a ``Vary: Cookie``
    header, which allows caching proxies to cache different pages for
    different users.

    Default is ``False``.

    .. attribute:: permanent

    This sets and reflects the ``'_permanent'`` key in the dict.

    Default is ``False``.

    """

    def __bool__(self) -> bool:
        return bool(dict(self)) and self.keys() != {"_permanent"}

    def __init__(
        self,
        initial: Optional[Dict[str, Any]] = None,
        sid: Optional[str] = None,
        permanent: Optional[bool] = None,
    ):
        def on_update(self) -> None:
            self.modified = True
            self.accessed = True

        CallbackDict.__init__(self, initial, on_update)
        self.sid = sid
        if permanent:
            self.permanent = permanent
        self.modified = False
        self.accessed = False

    def __getitem__(self, key: str) -> Any:
        self.accessed = True
        return super().__getitem__(key)

    def get(self, key: str, default: Any = None) -> Any:
        self.accessed = True
        return super().get(key, default)

    def setdefault(self, key: str, default: Any = None) -> Any:
        self.accessed = True
        return super().setdefault(key, default)

    def clear(self) -> None:
        """Clear the session except for the '_permanent' key."""
        permanent = self.get("_permanent", False)
        super().clear()
        self["_permanent"] = permanent


class Serializer(ABC):
    """Baseclass for session serialization."""

    @abstractmethod
    def decode(self, serialized_data: bytes) -> dict:
        """Deserialize the session data."""
        raise NotImplementedError()

    @abstractmethod
    def encode(self, session: ServerSideSession) -> bytes:
        """Serialize the session data."""
        raise NotImplementedError()


class MsgSpecSerializer(Serializer):
    def __init__(self, app: Flask, format: str):
        self.app: Flask = app
        self.encoder: msgspec.msgpack.Encoder or msgspec.json.Encoder
        self.decoder: msgspec.msgpack.Decoder or msgspec.json.Decoder
        self.alternate_decoder: msgspec.msgpack.Decoder or msgspec.json.Decoder

        if format == "msgpack":
            self.encoder = msgspec.msgpack.Encoder()
            self.decoder = msgspec.msgpack.Decoder()
            self.alternate_decoder = msgspec.json.Decoder()
        elif format == "json":
            self.encoder = msgspec.json.Encoder()
            self.decoder = msgspec.json.Decoder()
            self.alternate_decoder = msgspec.msgpack.Decoder()
        else:
            raise ValueError(f"Unsupported serialization format: {format}")

    def encode(self, session: ServerSideSession) -> bytes:
        """Serialize the session data."""
        try:
            return self.encoder.encode(dict(session))
        except Exception as e:
            self.app.logger.error(f"Failed to serialize session data: {e}")
            raise

    def decode(self, serialized_data: bytes) -> dict:
        """Deserialize the session data."""
        # TODO: Remove the pickle fallback in 1.0.0
        with suppress(msgspec.DecodeError):
            return self.decoder.decode(serialized_data)
        with suppress(msgspec.DecodeError):
            return self.alternate_decoder.decode(serialized_data)
        with suppress(pickle.UnpicklingError):
            return pickle.loads(serialized_data)
        # If all decoders fail, raise the final exception
        self.app.logger.error("Failed to deserialize session data", exc_info=True)
        raise pickle.UnpicklingError("Failed to deserialize session data")


class ServerSideSessionInterface(FlaskSessionInterface, ABC):
    """Used to open a :class:`flask.sessions.ServerSideSessionInterface` instance."""

    session_class = ServerSideSession
    serializer = None
    ttl = True

    def __init__(
        self,
        app: Flask,
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
        cleanup_n_requests: Optional[int] = Defaults.SESSION_CLEANUP_N_REQUESTS,
    ):
        self.app = app
        self.key_prefix = key_prefix
        self.use_signer = use_signer
        if use_signer:
            warnings.warn(
                "The 'use_signer' option is deprecated and will be removed in the next minor release. "
                "Please update your configuration accordingly or open an issue.",
                DeprecationWarning,
                stacklevel=1,
            )
        self.permanent = permanent
        self.sid_length = sid_length
        self.has_same_site_capability = hasattr(self, "get_cookie_samesite")
        self.cleanup_n_requests = cleanup_n_requests

        # Cleanup settings for non-TTL databases only
        if getattr(self, "ttl", None) is False:
            if self.cleanup_n_requests:
                self.app.before_request(self._cleanup_n_requests)
            else:
                self._register_cleanup_app_command()

        # Set the serialization format
        self.serializer = MsgSpecSerializer(format=serialization_format, app=app)

    # INTERNAL METHODS

    def _generate_sid(self, session_id_length: int) -> str:
        """Generate a random session id."""
        return secrets.token_urlsafe(session_id_length)

    # TODO: Remove in 1.0.0
    def _get_signer(self, app: Flask) -> Signer:
        if not hasattr(app, "secret_key") or not app.secret_key:
            raise KeyError("SECRET_KEY must be set when SESSION_USE_SIGNER=True")
        return Signer(app.secret_key, salt="flask-session", key_derivation="hmac")

    # TODO: Remove in 1.0.0
    def _unsign(self, app, sid: str) -> str:
        signer = self._get_signer(app)
        sid_as_bytes = signer.unsign(sid)
        sid = sid_as_bytes.decode()
        return sid

    # TODO: Remove in 1.0.0
    def _sign(self, app, sid: str) -> str:
        signer = self._get_signer(app)
        sid_as_bytes = want_bytes(sid)
        return signer.sign(sid_as_bytes).decode("utf-8")

    def _get_store_id(self, sid: str) -> str:
        return self.key_prefix + sid

    def should_set_storage(self, app: Flask, session: ServerSideSession) -> bool:
        """Used by session backends to determine if session in storage
        should be set for this session cookie for this response. If the session
        has been modified, the session is set to storage. If
        the ``SESSION_REFRESH_EACH_REQUEST`` config is true, the session is
        always set to storage. In the second case, this means refreshing the
        storage expiry even if the session has not been modified.

        .. versionadded:: 0.7.0
        """

        return session.modified or app.config["SESSION_REFRESH_EACH_REQUEST"]

    # CLEANUP METHODS FOR NON TTL DATABASES

    def _register_cleanup_app_command(self):
        """
        Register a custom Flask CLI command for cleaning up expired sessions.

        Run the command with `flask session_cleanup`. Run with a cron job
        or scheduler such as Heroku Scheduler to automatically clean up expired sessions.
        """

        @self.app.cli.command("session_cleanup")
        def session_cleanup():
            with self.app.app_context():
                self._delete_expired_sessions()

    def _cleanup_n_requests(self) -> None:
        """
        Delete expired sessions on average every N requests.

        This is less desirable than using the scheduled app command cleanup as it may
        slow down some requests but may be useful for rapid development.
        """
        if self.cleanup_n_requests and random.randint(0, self.cleanup_n_requests) == 0:
            self._delete_expired_sessions()

    # SECURITY API METHODS

    def regenerate(self, session: ServerSideSession) -> None:
        """Regenerate the session id for the given session. Can be used by calling ``flask.session_interface.regenerate()``."""
        if session:
            # Remove the old session from storage
            self._delete_session(self._get_store_id(session.sid))
            # Generate a new session ID
            new_sid = self._generate_sid(self.sid_length)
            session.sid = new_sid
            # Mark the session as modified to ensure it gets saved
            session.modified = True

    # METHODS OVERRIDE FLASK SESSION INTERFACE

    def save_session(
        self, app: Flask, session: ServerSideSession, response: Response
    ) -> None:

        # Get the domain and path for the cookie from the app
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        name = self.get_cookie_name(app)

        # Generate a prefixed session id
        store_id = self._get_store_id(session.sid)

        # Add a "Vary: Cookie" header if the session was accessed at all.
        # This assumes the app is checking the session values in a request that
        # behaves differently based on those values. ie. session.get("is_authenticated")
        if session.accessed:
            response.vary.add("Cookie")

        # If the session is empty, do not save it to the database or set a cookie
        if not session:
            # If the session was deleted (empty and modified), delete the saved session  from the database and tell the client to delete the cookie
            if session.modified:
                self._delete_session(store_id)
                response.delete_cookie(key=name, domain=domain, path=path)
                response.vary.add("Cookie")
            return

        if not self.should_set_storage(app, session):
            return

        # Update existing or create new session in the database
        self._upsert_session(app.permanent_session_lifetime, session, store_id)

        if not self.should_set_cookie(app, session):
            return

        # Get the additional required cookie settings
        value = self._sign(app, session.sid) if self.use_signer else session.sid
        expires = self.get_expiration_time(app, session)
        httponly = self.get_cookie_httponly(app)
        secure = self.get_cookie_secure(app)
        samesite = (
            self.get_cookie_samesite(app) if self.has_same_site_capability else None
        )

        # Set the browser cookie
        response.set_cookie(
            key=name,
            value=value,
            expires=expires,
            httponly=httponly,
            domain=domain,
            path=path,
            secure=secure,
            samesite=samesite,
        )
        response.vary.add("Cookie")

    def open_session(self, app: Flask, request: Request) -> ServerSideSession:
        # Get the session ID from the cookie
        sid = request.cookies.get(app.config["SESSION_COOKIE_NAME"])

        # If there's no session ID, generate a new one
        if not sid:
            sid = self._generate_sid(self.sid_length)
            return self.session_class(sid=sid, permanent=self.permanent)
        # If the session ID is signed, unsign it
        if self.use_signer:
            try:
                sid = self._unsign(app, sid)
            except BadSignature:
                sid = self._generate_sid(self.sid_length)
                return self.session_class(sid=sid, permanent=self.permanent)

        # Retrieve the session data from the database
        store_id = self._get_store_id(sid)
        saved_session_data = self._retrieve_session_data(store_id)

        # If the saved session exists, load the session data from the document
        if saved_session_data is not None:
            return self.session_class(saved_session_data, sid=sid)

        # If the saved session does not exist, create a new session
        sid = self._generate_sid(self.sid_length)
        return self.session_class(sid=sid, permanent=self.permanent)

    # METHODS TO BE IMPLEMENTED BY SUBCLASSES

    @abstractmethod
    @retry_query()  # use only when retry not supported directly by the client
    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        """Get the saved session from the session storage."""
        raise NotImplementedError()

    @abstractmethod
    @retry_query()  # use only when retry not supported directly by the client
    def _delete_session(self, store_id: str) -> None:
        """Delete session from the session storage."""
        raise NotImplementedError()

    @abstractmethod
    @retry_query()  # use only when retry not supported directly by the client
    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        """Update existing or create new session in the session storage."""
        raise NotImplementedError()

    @retry_query()  # use only when retry not supported directly by the client
    def _delete_expired_sessions(self) -> None:
        """Delete expired sessions from the session storage. Only required for non-TTL databases."""
        pass
